//
// clamodo: clamav/qmail-queue integration tool
//

// Copyright (c) 2004 Dean Gaudet <dean@arctic.org>
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.

// $Id: clamodo.c,v 1.2 2004/10/25 22:27:27 dean Exp $

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>

#define QMAILQUEUE	"/var/qmail/bin/qmail-queue"
#define CLAMDSCAN	"/usr/bin/clamdscan"
#define TMP		"/tmp"

// qmail-queue exit codes
//
// note that qmail-1.03/qmail.c doesn't normally give a nice
// virus response message to exit 33 (although it does consider exit 33
// to be a permanent failure).  you'll either want to change VIRUS_FAIL
// to 31, or add something like this to the switch in your qmail.c:
//
//   case 33: return "Dmessage rejected due to detected virus (#5.3.0)";
//
#define VIRUS_FAIL	(33)
#define TEMP_FAIL	(71)

// you shouldn't need to configure anything beyond here

static pid_t ppid;
static int unlink_msg;
static char msg_filename[] = "/tmp/clamodo.XXXXXX";
static int unlink_scanerr;
static char scanerr_filename[] = "/tmp/clamodo.XXXXXX";

static void die(const char *msg)
{
	fprintf(stderr, "clamodo ppid %d %s: %s\n", ppid, msg, strerror(errno));
	if (unlink_msg) unlink(msg_filename);
	if (unlink_scanerr) unlink(scanerr_filename);
	exit(TEMP_FAIL);
}


static int copy_fd(int from_fd, int to_fd)
{
	char buf[4096];
	ssize_t rc;
	ssize_t amt;
	char *p;

	for (;;) {
		rc = read(from_fd, buf, sizeof(buf));
		if (rc == 0) return 0;
		if (rc < 0) {
			if (errno == EINTR) continue;
			return rc;
		}
		amt = rc;
		p = buf;
		while (amt) {
			rc = write(to_fd, p, amt);
			if (rc < 0) {
				if (errno != EINTR) return rc;
			}
			else {
				p += rc;
				amt -= rc;
			}
		}
	}
}


int main(int argc, char **argv)
{
	int msg_fd;
	int scanerr_fd;
	pid_t child;
	int child_status;

	ppid = getppid();
	signal(SIGPIPE, SIG_IGN);

	// we really have two design choices here -- we could open
	// up a pipe to both clamdscan and qmail-queue and relay
	// the message to both simultaneously.  however, clamdscan
	// is just going to write it to disk anyhow... so we might
	// as well write it to disk ourselves, and invoke the
	// two processes serially.
	if ((msg_fd = mkstemp(msg_filename)) == -1) die("mkstemp() failed");
	unlink_msg = 1;
	if (fchmod(msg_fd, 0640) == -1) die("fchmod() failed");
	if (copy_fd(0, msg_fd) == -1) die("copy_fd() failed");
	if (close(msg_fd) == -1) die("close(msg_fd) failed");

	// done with fd 0, but open /dev/null to prevent mistakes
	close(0); if (open("/dev/null", O_RDONLY) != 0) die("unable to open /dev/null");

	// we will capture the stderr output of clamdscan, and dump it
	// to our stderr (prefixing each line) when a virus is detected
	// or error occurs.
	if ((scanerr_fd = mkstemp(scanerr_filename)) == -1) die("mkstemp() failed");
	unlink_scanerr = 1;

	child = fork();
	if (child == -1) die("fork() failed");
	else if (child == 0) {
		if (dup2(scanerr_fd, 2) == -1) die("dup2(scanerr_fd, 2) failed");
		close(1); if (open("/dev/null", O_RDONLY) != 1) die("unable to open /dev/null");
		execl(CLAMDSCAN, CLAMDSCAN, "--no-summary", msg_filename, NULL);
		die("exec(" CLAMDSCAN ") failed");
	}
	close(scanerr_fd);
	if (waitpid(child, &child_status, 0) == -1) die("waitpid(clamdscan) failed");
	if (!WIFEXITED(child_status) || WEXITSTATUS(child_status) != 0) {
		FILE *scanerr;

		scanerr = fopen(scanerr_filename, "r");
		if (scanerr) {
			int issue_prefix = 1;
			int c;

			while ((c = fgetc(scanerr)) != EOF) {
				if (issue_prefix) {
					fprintf(stderr, "clamodo ppid %d ", ppid);
					issue_prefix = 0;
				}
				fputc(c, stderr);
				if (c == '\n') issue_prefix = 1;
			}
			fclose(scanerr);
		}
		unlink(msg_filename); unlink_msg = 0;
		unlink(scanerr_filename); unlink_scanerr = 0;
		if (!WIFEXITED(child_status)) {
			fprintf(stderr, "clamodo ppid %d clamdscan abnormal exit: %d\n", ppid, child_status);
			exit(TEMP_FAIL);
		}
		exit(WEXITSTATUS(child_status) == 1 ? VIRUS_FAIL : TEMP_FAIL);
	}
	unlink(scanerr_filename); unlink_scanerr = 0;

	msg_fd = open(msg_filename, O_RDONLY);
	if (msg_fd == -1) die("unable to re-open msg");
	unlink(msg_filename); unlink_msg = 0;
	if (dup2(msg_fd, 0) == -1) die("dup2(msg_fd, 0) failed");
	execl(QMAILQUEUE, QMAILQUEUE, NULL);
	die("exec(" QMAILQUEUE ") failed");
	return TEMP_FAIL;
}
